--========================================================================
-- GLFW - An OpenGL framework
-- File:        glfw.e
-- API version: 2.4
-- Author:      Marcus Geelnard (marcus.geelnard at home.se)
-- WWW:         http://glfw.sourceforge.net
--------------------------------------------------------------------------
-- Copyright (c) 2002-2003 Marcus Geelnard
--
-- GLFW is provided "as-is", without any express or implied warranty,
-- without even the implied warranty of fitness for a particular purpose.
-- In no event will the author be held liable for any damage araising from
-- the use of GLFW.
--
-- Permission is granted to anyone to use GLFW for any purpose, including
-- commercial applications, and to alter it and redistribute it freely,
-- subject to the following restrictions:
--
-- 1. You must not claim that you wrote GLFW.
-- 2. If you modify any parts of GLFW, it must be clearly stated which
--    modifications have been made, and that those modifications are not
--    the work of the author of GLFW.
-- 3. If you use GLFW in a product, an acknowledgment in the product
--    documentation would be appreciated, but is not required.
-- 4. This notice may not be removed or altered from any source
--    distribution of GLFW.
--
-- Marcus Geelnard
-- marcus.geelnard at home.se

include dll.e
include machine.e
atom GlfwDll
if platform() = 2 then
	GlfwDll = open_dll("GLFW.dll")
elsif platform() = 3 then
	GlfwDll = open_dll("libglfw.so")
end if

--========================================================================
-- GLFW version
--========================================================================

global constant GLFW_VERSION_MAJOR = 2,
				GLFW_VERSION_MINOR = 4,
				GLFW_VERSION_REVISION = 0

--========================================================================
-- Input handling definitions
--========================================================================

-- Key and button state/action definitions
global constant GLFW_RELEASE = 0,
				GLFW_PRESS = 1

-- Keyboard key definitions: 8-bit ISO-8859-1 (Latin 1) encoding is used
-- for printable keys (such as A-Z, 0-9 etc), and values above 256
-- represent special (non-printable) keys (e.g. F1, Page Up etc).
global constant GLFW_KEY_UNKNOWN = -1,
				GLFW_KEY_SPACE = 32,
				GLFW_KEY_SPECIAL = 256,
				GLFW_KEY_ESC = (GLFW_KEY_SPECIAL+1),
				GLFW_KEY_F1 = (GLFW_KEY_SPECIAL+2),
				GLFW_KEY_F2 = (GLFW_KEY_SPECIAL+3),
				GLFW_KEY_F3 = (GLFW_KEY_SPECIAL+4),
				GLFW_KEY_F4 = (GLFW_KEY_SPECIAL+5),
				GLFW_KEY_F5 = (GLFW_KEY_SPECIAL+6),
				GLFW_KEY_F6 = (GLFW_KEY_SPECIAL+7),
				GLFW_KEY_F7 = (GLFW_KEY_SPECIAL+8),
				GLFW_KEY_F8 = (GLFW_KEY_SPECIAL+9),
				GLFW_KEY_F9 = (GLFW_KEY_SPECIAL+10),
				GLFW_KEY_F10 = (GLFW_KEY_SPECIAL+11),
				GLFW_KEY_F11 = (GLFW_KEY_SPECIAL+12),
				GLFW_KEY_F12 = (GLFW_KEY_SPECIAL+13),
				GLFW_KEY_F13 = (GLFW_KEY_SPECIAL+14),
				GLFW_KEY_F14 = (GLFW_KEY_SPECIAL+15),
				GLFW_KEY_F15 = (GLFW_KEY_SPECIAL+16),
				GLFW_KEY_F16 = (GLFW_KEY_SPECIAL+17),
				GLFW_KEY_F17 = (GLFW_KEY_SPECIAL+18),
				GLFW_KEY_F18 = (GLFW_KEY_SPECIAL+19),
				GLFW_KEY_F19 = (GLFW_KEY_SPECIAL+20),
				GLFW_KEY_F20 = (GLFW_KEY_SPECIAL+21),
				GLFW_KEY_F21 = (GLFW_KEY_SPECIAL+22),
				GLFW_KEY_F22 = (GLFW_KEY_SPECIAL+23),
				GLFW_KEY_F23 = (GLFW_KEY_SPECIAL+24),
				GLFW_KEY_F24 = (GLFW_KEY_SPECIAL+25),
				GLFW_KEY_F25 = (GLFW_KEY_SPECIAL+26),
				GLFW_KEY_UP = (GLFW_KEY_SPECIAL+27),
				GLFW_KEY_DOWN = (GLFW_KEY_SPECIAL+28),
				GLFW_KEY_LEFT = (GLFW_KEY_SPECIAL+29),
				GLFW_KEY_RIGHT = (GLFW_KEY_SPECIAL+30),
				GLFW_KEY_LSHIFT = (GLFW_KEY_SPECIAL+31),
				GLFW_KEY_RSHIFT = (GLFW_KEY_SPECIAL+32),
				GLFW_KEY_LCTRL = (GLFW_KEY_SPECIAL+33),
				GLFW_KEY_RCTRL = (GLFW_KEY_SPECIAL+34),
				GLFW_KEY_LALT = (GLFW_KEY_SPECIAL+35),
				GLFW_KEY_RALT = (GLFW_KEY_SPECIAL+36),
				GLFW_KEY_TAB = (GLFW_KEY_SPECIAL+37),
				GLFW_KEY_ENTER = (GLFW_KEY_SPECIAL+38),
				GLFW_KEY_BACKSPACE = (GLFW_KEY_SPECIAL+39),
				GLFW_KEY_INSERT = (GLFW_KEY_SPECIAL+40),
				GLFW_KEY_DEL = (GLFW_KEY_SPECIAL+41),
				GLFW_KEY_PAGEUP = (GLFW_KEY_SPECIAL+42),
				GLFW_KEY_PAGEDOWN = (GLFW_KEY_SPECIAL+43),
				GLFW_KEY_HOME = (GLFW_KEY_SPECIAL+44),
				GLFW_KEY_END = (GLFW_KEY_SPECIAL+45),
				GLFW_KEY_KP_0 = (GLFW_KEY_SPECIAL+46),
				GLFW_KEY_KP_1 = (GLFW_KEY_SPECIAL+47),
				GLFW_KEY_KP_2 = (GLFW_KEY_SPECIAL+48),
				GLFW_KEY_KP_3 = (GLFW_KEY_SPECIAL+49),
				GLFW_KEY_KP_4 = (GLFW_KEY_SPECIAL+50),
				GLFW_KEY_KP_5 = (GLFW_KEY_SPECIAL+51),
				GLFW_KEY_KP_6 = (GLFW_KEY_SPECIAL+52),
				GLFW_KEY_KP_7 = (GLFW_KEY_SPECIAL+53),
				GLFW_KEY_KP_8 = (GLFW_KEY_SPECIAL+54),
				GLFW_KEY_KP_9 = (GLFW_KEY_SPECIAL+55),
				GLFW_KEY_KP_DIVIDE = (GLFW_KEY_SPECIAL+56),
				GLFW_KEY_KP_MULTIPLY = (GLFW_KEY_SPECIAL+57),
				GLFW_KEY_KP_SUBTRACT = (GLFW_KEY_SPECIAL+58),
				GLFW_KEY_KP_ADD = (GLFW_KEY_SPECIAL+59),
				GLFW_KEY_KP_DECIMAL = (GLFW_KEY_SPECIAL+60),
				GLFW_KEY_KP_EQUAL = (GLFW_KEY_SPECIAL+61),
				GLFW_KEY_KP_ENTER = (GLFW_KEY_SPECIAL+62),
				GLFW_KEY_LAST = GLFW_KEY_KP_ENTER

-- Mouse button definitions
global constant GLFW_MOUSE_BUTTON_LEFT = 0,
				GLFW_MOUSE_BUTTON_RIGHT = 1,
				GLFW_MOUSE_BUTTON_MIDDLE = 2,
				GLFW_MOUSE_BUTTON_LAST = GLFW_MOUSE_BUTTON_MIDDLE

-- Joystick identifiers
global constant GLFW_JOYSTICK_1 = 0,
				GLFW_JOYSTICK_2 = 1,
				GLFW_JOYSTICK_3 = 2,
				GLFW_JOYSTICK_4 = 3,
				GLFW_JOYSTICK_5 = 4,
				GLFW_JOYSTICK_6 = 5,
				GLFW_JOYSTICK_7 = 6,
				GLFW_JOYSTICK_8 = 7,
				GLFW_JOYSTICK_9 = 8,
				GLFW_JOYSTICK_10 = 9,
				GLFW_JOYSTICK_11 = 10,
				GLFW_JOYSTICK_12 = 11,
				GLFW_JOYSTICK_13 = 12,
				GLFW_JOYSTICK_14 = 13,
				GLFW_JOYSTICK_15 = 14,
				GLFW_JOYSTICK_16 = 15,
				GLFW_JOYSTICK_LAST = GLFW_JOYSTICK_16

--========================================================================
-- Other definitions
--========================================================================

-- glfwOpenWindow modes
global constant GLFW_WINDOW = #00010001,
				GLFW_FULLSCREEN  = #00010002

-- glfwGetWindowParam tokens
global constant GLFW_OPENED = #00020001,
				GLFW_ACTIVE = #00020002,
				GLFW_ICONIFIED = #00020003,
				GLFW_ACCELERATED = #00020004,
				GLFW_RED_BITS = #00020005,
				GLFW_GREEN_BITS = #00020006,
				GLFW_BLUE_BITS = #00020007,
				GLFW_ALPHA_BITS = #00020008,
				GLFW_DEPTH_BITS = #00020009,
				GLFW_STENCIL_BITS = #0002000A,

-- The following constants are used for both glfwGetWindowParam
-- and glfwOpenWindowHint
				GLFW_REFRESH_RATE = #0002000B,
				GLFW_ACCUM_RED_BITS = #0002000C,
				GLFW_ACCUM_GREEN_BITS = #0002000D,
				GLFW_ACCUM_BLUE_BITS = #0002000E,
				GLFW_ACCUM_ALPHA_BITS = #0002000F,
				GLFW_AUX_BUFFERS = #00020010,
				GLFW_STEREO = #00020011,

-- glfwEnable/glfwDisable tokens
				GLFW_MOUSE_CURSOR = #00030001,
				GLFW_STICKY_KEYS = #00030002,
				GLFW_STICKY_MOUSE_BUTTONS = #00030003,
				GLFW_SYSTEM_KEYS = #00030004,
				GLFW_KEY_REPEAT = #00030005,
				GLFW_AUTO_POLL_EVENTS = #00030006,

-- glfwWaitThread wait modes
				GLFW_WAIT = #00040001,
				GLFW_NOWAIT = #00040002,

-- glfwGetJoystickParam tokens
				GLFW_PRESENT = #00050001,
				GLFW_AXES = #00050002,
				GLFW_BUTTONS = #00050003

-- glfwReadImage/glfwLoadTexture2D flags
global constant GLFW_NO_RESCALE_BIT = #00000001,	-- Only for glfwReadImage
				GLFW_ORIGIN_UL_BIT = #00000002,
				GLFW_BUILD_MIPMAPS_BIT = #00000004 -- Only for glfwLoadTexture2D

-- Time spans longer than this (seconds) are considered to be infinity
global constant GLFW_INFINITY = 100000.0


--========================================================================
-- Typedefs
--========================================================================

-- The video mode structure used by glfwGetVideoModes()
global type GLFWvidmode(object gt)
	return sequence(gt)
end type
global constant
		GlfwVmWidth = 1,
		GlfwVmHeight = 2,
		GlfwVmRedBits = 3,
		GlfwVmBlueBits = 4,
		GlfwVmGreenBits = 5

-- Image/texture information
global type GLFWimage(object gt)
	return atom(gt)
end type
global constant
		GlfwIWidth = 0,
		GlfwIHeight = 4,
		GlfwIFormat = 8,
		GlfwIBytesPerPixel = 12,
		GlfwIData = 16,
		szGLFWimage = 20

-- Thread ID
global type GLFWthread(object gt)
	--int
	return atom(gt)
end type

-- Mutex object
global type GLFWmutex(object gt)
	return atom(gt)
end type

-- Condition variable object
global type GLFWcond(object gt)
	return atom(gt)
end type

-- Function pointer types
--procedure GLFWwindowsizefun(integer,integer)
--procedure GLFWmousebuttonfun(integer,integer)
--procedure GLFWmouseposfun(integer,integer)
--procedure GLFWmousewheelfun(integer)
--procedure GLFWkeyfun(integer,integer)
--procedure GLFWcharfun(integer,integer)
--procedure GLFWthreadfun(atom)

--========================================================================
-- Prototypes
--========================================================================
constant
GlfwInit = define_c_func(GlfwDll, "glfwInit", {}, C_INT),
GlfwTerminate = define_c_proc(GlfwDll, "glfwTerminate", {}),
GlfwGetVersion = define_c_proc(GlfwDll, "glfwGetVersion", {C_POINTER, C_POINTER, C_POINTER}),
GlfwOpenWindow = define_c_func(GlfwDll, "glfwOpenWindow", {C_INT, C_INT, C_INT, C_INT, C_INT, C_INT, C_INT, C_INT, C_INT}, C_INT),
GlfwOpenWindowHint = define_c_proc(GlfwDll, "glfwOpenWindowHint", {C_INT, C_INT}),
GlfwCloseWindow = define_c_proc(GlfwDll, "glfwCloseWindow", {}),
GlfwSetWindowTitle = define_c_proc(GlfwDll, "glfwSetWindowTitle", {C_POINTER}),
GlfwGetWindowSize = define_c_proc(GlfwDll, "glfwGetWindowSize", {C_POINTER, C_POINTER}),
GlfwSetWindowSize = define_c_proc(GlfwDll, "glfwSetWindowSize", {C_INT, C_INT}),
GlfwSetWindowPos = define_c_proc(GlfwDll, "glfwSetWindowPos", {C_INT, C_INT}),
GlfwIconifyWindow = define_c_proc(GlfwDll, "glfwIconifyWindow", {}),
GlfwRestoreWindow = define_c_proc(GlfwDll, "glfwRestoreWindow", {}),
GlfwSwapBuffers = define_c_proc(GlfwDll, "glfwSwapBuffers", {}),
GlfwSwapInterval = define_c_proc(GlfwDll, "glfwSwapInterval", {C_INT}),
GlfwGetWindowParam = define_c_func(GlfwDll, "glfwGetWindowParam", {C_INT}, C_INT),
GlfwSetWindowSizeCallback = define_c_proc(GlfwDll, "glfwSetWindowSizeCallback", {C_POINTER}),
GlfwGetVideoModes = define_c_func(GlfwDll, "glfwGetVideoModes", {C_POINTER, C_INT}, C_INT),
GlfwGetDesktopMode = define_c_proc(GlfwDll, "glfwGetDesktopMode", {C_POINTER}),
GlfwPollEvents = define_c_proc(GlfwDll, "glfwPollEvents", {}),
GlfwGetKey = define_c_func(GlfwDll, "glfwGetKey", {C_INT}, C_INT),
GlfwGetMouseButton = define_c_func(GlfwDll, "glfwGetMouseButton", {C_INT}, C_INT),
GlfwGetMousePos = define_c_proc(GlfwDll, "glfwGetMousePos", {C_POINTER, C_POINTER}),
GlfwSetMousePos = define_c_proc(GlfwDll, "glfwSetMousePos", {C_INT, C_INT}),
GlfwGetMouseWheel = define_c_func(GlfwDll, "glfwGetMouseWheel", {}, C_INT),
GlfwSetMouseWheel = define_c_proc(GlfwDll, "glfwSetMouseWheel", {C_INT}),
GlfwSetKeyCallback = define_c_proc(GlfwDll, "glfwSetKeyCallback", {C_POINTER}),
GlfwSetCharCallback = define_c_proc(GlfwDll, "glfwSetCharCallback", {C_POINTER}),
GlfwSetMouseButtonCallback = define_c_proc(GlfwDll, "glfwSetMouseButtonCallback", {C_POINTER}),
GlfwSetMousePosCallback = define_c_proc(GlfwDll, "glfwSetMousePosCallback", {C_POINTER}),
GlfwSetMouseWheelCallback = define_c_proc(GlfwDll, "glfwSetMouseWheelCallback", {C_POINTER}),
GlfwGetJoystickParam = define_c_func(GlfwDll, "glfwGetJoystickParam", {C_INT, C_INT}, C_INT),
GlfwGetJoystickPos = define_c_func(GlfwDll, "glfwGetJoystickPos", {C_INT, C_POINTER, C_INT}, C_INT),
GlfwGetJoystickButtons = define_c_func(GlfwDll, "glfwGetJoystickButtons", {C_INT, C_POINTER, C_INT}, C_INT),
GlfwGetTime = define_c_func(GlfwDll, "glfwGetTime", {}, C_DOUBLE),
GlfwSetTime = define_c_proc(GlfwDll, "glfwSetTime", {C_DOUBLE}),
GlfwSleep = define_c_proc(GlfwDll, "glfwSleep", {C_DOUBLE}),
GlfwExtensionSupported = define_c_func(GlfwDll, "glfwExtensionSupported", {C_POINTER}, C_INT),
GlfwGetProcAddress = define_c_func(GlfwDll, "glfwGetProcAddress", {C_POINTER}, C_POINTER),
GlfwGetGLVersion = define_c_proc(GlfwDll, "glfwGetGLVersion", {C_POINTER, C_POINTER, C_POINTER}),
GlfwCreateThread = define_c_func(GlfwDll, "glfwCreateThread", {C_POINTER, C_POINTER}, C_INT),
GlfwDestroyThread = define_c_proc(GlfwDll, "glfwDestroyThread", {C_INT}),
GlfwWaitThread = define_c_func(GlfwDll, "glfwWaitThread", {C_INT, C_INT}, C_INT),
GlfwGetThreadID = define_c_func(GlfwDll, "glfwGetThreadID", {}, C_INT),
GlfwCreateMutex = define_c_func(GlfwDll, "glfwCreateMutex", {}, C_POINTER),
GlfwDestroyMutex = define_c_proc(GlfwDll, "glfwDestroyMutex", {C_POINTER}),
GlfwLockMutex = define_c_proc(GlfwDll, "glfwLockMutex", {C_POINTER}),
GlfwUnlockMutex = define_c_proc(GlfwDll, "glfwUnlockMutex", {C_POINTER}),
GlfwCreateCond = define_c_func(GlfwDll, "glfwCreateCond", {}, C_POINTER),
GlfwDestroyCond = define_c_proc(GlfwDll, "glfwDestroyCond", {C_POINTER}),
GlfwWaitCond = define_c_proc(GlfwDll, "glfwWaitCond", {C_POINTER, C_POINTER, C_DOUBLE}),
GlfwSignalCond = define_c_proc(GlfwDll, "glfwSignalCond", {C_POINTER}),
GlfwBroadcastCond = define_c_proc(GlfwDll, "glfwBroadcastCond", {C_POINTER}),
GlfwGetNumberOfProcessors = define_c_func(GlfwDll, "glfwGetNumberOfProcessors", {}, C_INT),
GlfwEnable = define_c_proc(GlfwDll, "glfwEnable", {C_INT}),
GlfwDisable = define_c_proc(GlfwDll, "glfwDisable", {C_INT}),
GlfwReadImage = define_c_func(GlfwDll, "glfwReadImage", {C_POINTER, C_POINTER, C_INT}, C_INT),
GlfwFreeImage = define_c_proc(GlfwDll, "glfwFreeImage", {C_POINTER}),
GlfwLoadTexture2D = define_c_func(GlfwDll, "glfwLoadTexture2D", {C_POINTER, C_INT}, C_INT)

-- GLFW initialization, termination and version querying
global function glfwInit()
	return c_func(GlfwInit, {})
end function

global procedure glfwTerminate()
	c_proc(GlfwTerminate, {})
end procedure

global function glfwGetVersion()
atom lpver
sequence version
lpver = allocate(12)
c_proc(GlfwGetVersion, {lpver, lpver + 4, lpver + 8})
version = peek4s({lpver, 3})
free(lpver)
	return version
end function

-- Window handling
global function glfwOpenWindow(integer width, integer height,
				integer redbits, integer greenbits, integer bluebits, integer alphabits,
				integer depthbits, integer stencilbits, integer mode)
	return c_func(GlfwOpenWindow, {width, height,
									redbits, greenbits, bluebits, alphabits,
									depthbits, stencilbits, mode})
end function

global procedure glfwOpenWindowHint(integer target, integer hint)
	c_proc(GlfwOpenWindowHint, {target, hint})
end procedure

global procedure glfwCloseWindow()
	c_proc(GlfwCloseWindow, {})
end procedure

global procedure glfwSetWindowTitle(object title)
	title = allocate_string(title)
	c_proc(GlfwSetWindowTitle, {title})
	free(title)
end procedure

global function glfwGetWindowSize()
atom lpsize
sequence size
lpsize = allocate(8)
c_proc(GlfwGetWindowSize, {lpsize, lpsize + 4})
size = peek4s({lpsize, 3})
free(lpsize)
	return size
end function

global procedure glfwSetWindowSize(integer width, integer height)
	c_proc(GlfwSetWindowSize, {width, height})
end procedure

global procedure glfwSetWindowPos(integer x, integer y)
	c_proc(GlfwSetWindowPos, {x, y})
end procedure

global procedure glfwIconifyWindow()
	c_proc(GlfwIconifyWindow, {})
end procedure

global procedure glfwRestoreWindow()
	c_proc(GlfwRestoreWindow, {})
end procedure

global procedure glfwSwapBuffers()
	c_proc(GlfwSwapBuffers, {})
end procedure

global procedure glfwSwapInterval(integer interval)
	c_proc(GlfwSwapInterval, {interval})
end procedure

global function glfwGetWindowParam(integer param)
	return c_func(GlfwGetWindowParam, {param})
end function

global procedure glfwSetWindowSizeCallback(atom cbfun)
	c_proc(GlfwSetWindowSizeCallback, {cbfun})
end procedure

-- Video mode functions
global function glfwGetVideoModes(integer maxcount)
integer ret
sequence modes
atom list
list = allocate(4 * maxcount)
ret = c_func(GlfwGetVideoModes, {list, maxcount})
modes = repeat(0, ret)
for loop = 1 to ret do
	modes[loop] = peek({list, 5})
	list += 20
end for
free(list - (ret * 20))
	return modes
end function

global function glfwGetDesktopMode()
GLFWvidmode mode
atom lpmode
lpmode = allocate(20)
c_proc(GlfwGetDesktopMode, {lpmode})
mode = peek4s({lpmode, 5})
free(lpmode)
	return mode
end function

-- Input handling
global procedure glfwPollEvents()
	c_proc(GlfwPollEvents, {})
end procedure

global function glfwGetKey(integer key)
	return c_func(GlfwGetKey, {key})
end function

global function glfwGetMouseButton(integer button)
	return c_func(GlfwGetMouseButton, {button})
end function

global function glfwGetMousePos()
atom lppos
sequence pos
lppos = allocate(8)
c_proc(GlfwGetMousePos, {lppos, lppos + 4})
pos = peek4s({lppos, 3})
free(lppos)
	return pos
end function

global procedure glfwSetMousePos(integer xpos, integer ypos)
	c_proc(GlfwSetMousePos, {xpos, ypos})
end procedure

global function glfwGetMouseWheel()
	return c_func(GlfwGetMouseWheel, {})
end function

global procedure glfwSetMouseWheel(integer pos)
	c_proc(GlfwSetMouseWheel, {pos})
end procedure

global procedure glfwSetKeyCallback(atom cbfun)
	c_proc(GlfwSetKeyCallback, {cbfun})
end procedure

global procedure glfwSetCharCallback(atom cbfun)
	c_proc(GlfwSetCharCallback, {cbfun})
end procedure

global procedure glfwSetMouseButtonCallback(atom cbfun)
	c_proc(GlfwSetMouseButtonCallback, {cbfun})
end procedure

global procedure glfwSetMousePosCallback(atom cbfun)
	c_proc(GlfwSetMousePosCallback, {cbfun})
end procedure

global procedure glfwSetMouseWheelCallback(atom cbfun)
	c_proc(GlfwSetMouseWheelCallback, {cbfun})
end procedure

-- Joystick input
global function glfwGetJoystickParam(integer joy, integer param)
	return c_func(GlfwGetJoystickParam, {joy, param})
end function

global function glfwGetJoystickPos(integer joy, integer numaxes)
integer ret
sequence pos
atom lppos
lppos = allocate(4 * numaxes)
ret = c_func(GlfwGetJoystickPos, {joy, lppos, numaxes})
pos = repeat(0, ret)
for loop = 1 to ret do
	pos[loop] = float32_to_atom(peek({lppos, 4}))
	lppos += 4
end for
free(lppos - (ret * 4))
	return pos
end function

global function glfwGetJoystickButtons(integer joy, integer numbuttons)
integer ret
sequence buttons
atom lpbut
lpbut = allocate(4 * numbuttons)
ret = c_func(GlfwGetJoystickButtons, {joy, lpbut, numbuttons})
buttons = peek4s({lpbut, ret})
free(lpbut)
	return buttons
end function

-- Time
global function glfwGetTime()
	return c_func(GlfwGetTime, {})
end function

global procedure glfwSetTime(atom stime)
	c_proc(GlfwSetTime, {stime})
end procedure

global procedure glfwSleep(atom stime)
	c_proc(GlfwSleep, {stime})
end procedure

-- Extension support
global function glfwExtensionSupported(object extension)
integer ret
extension = allocate_string(extension)
ret = c_func(GlfwExtensionSupported, {extension})
free(extension)
	return ret
end function

global function glfwGetProcAddress(object procname)
atom ret
procname = allocate_string(procname)
ret = c_func(GlfwGetProcAddress, {procname})
free(procname)
	return ret
end function

global function glfwGetGLVersion()
atom lpver
sequence version
lpver = allocate(12)
c_proc(GlfwGetGLVersion, {lpver, lpver + 4, lpver + 8})
version = peek4s({lpver, 3})
free(lpver)
	return version
end function

-- Threading support
global function glfwCreateThread(atom fun, atom arg)
	return c_func(GlfwCreateThread, {fun, arg})
end function

global procedure glfwDestroyThread(GLFWthread ID)
	c_proc(GlfwDestroyThread, {ID})
end procedure

global function glfwWaitThread(GLFWthread ID, integer waitmode)
	return c_func(GlfwWaitThread, {ID, waitmode})
end function

global function glfwGetThreadID()
	return c_func(GlfwGetThreadID, {})
end function

global function glfwCreateMutex()
	return c_func(GlfwCreateMutex, {})
end function

global procedure glfwDestroyMutex(GLFWmutex mutex)
	c_proc(GlfwDestroyMutex, {mutex})
end procedure

global procedure glfwLockMutex(GLFWmutex mutex)
	c_proc(GlfwLockMutex, {mutex})
end procedure

global procedure glfwUnlockMutex(GLFWmutex mutex)
	c_proc(GlfwUnlockMutex, {mutex})
end procedure

global function glfwCreateCond()
	return c_func(GlfwCreateCond, {})
end function

global procedure glfwDestroyCond(GLFWcond cond)
	c_proc(GlfwDestroyCond, {cond})
end procedure

global procedure glfwWaitCond(GLFWcond cond, GLFWmutex mutex, atom timeout)
	c_proc(GlfwWaitCond, {cond, mutex, timeout})
end procedure

global procedure glfwSignalCond(GLFWcond cond)
	c_proc(GlfwSignalCond, {cond})
end procedure

global procedure glfwBroadcastCond(GLFWcond cond)
	c_proc(GlfwBroadcastCond, {cond})
end procedure

global function glfwGetNumberOfProcessors()
	return c_func(GlfwGetNumberOfProcessors, {})
end function

-- Enable/disable functions
global procedure glfwEnable(integer token)
	c_proc(GlfwEnable, {token})
end procedure

global procedure glfwDisable(integer token)
	c_proc(GlfwDisable, {token})
end procedure

-- Image/texture I/O support
global function glfwReadImage(object name, integer flags)
atom img
integer ret
name = allocate_string(name)
img = allocate(szGLFWimage)
ret = c_func(GlfwReadImage, {name, img, flags})
free(name)
if ret then
	return img
else
	free(img)
	return ret
end if
end function

global procedure glfwFreeImage(GLFWimage img)
	c_proc(GlfwFreeImage, {img})
end procedure

global function glfwLoadTexture2D(object name, integer flags)
integer ret
name = allocate_string(name)
ret = c_func(GlfwLoadTexture2D, {name, flags})
free(name)
	return ret
end function
